#define _CRT_RAND_S
#include <iostream>
#include <iomanip>
#include <functional>
#include <vector>
#include <list>
#include <cstdlib>
#include <ctime>

#include "benchmark.h"

#include "comb_sort.cpp"
#include "icomb_sort.cpp"
#include "quick_sort.cpp"
#include "pquick_sort.cpp"
#include "pquick_sort2.cpp"
#include "merge_sort.cpp"
#include "pmerge_sort.cpp"
#include "pmerge_sort2.cpp"
#include "mmerge_sort.cpp"
#include "heap_sort1.cpp"
#include "heap_sort2.cpp"
#include "heap_sort3.cpp"
#include "radix_sort.cpp"
#include "msd_radix_sort.cpp"
#include "diff_sort.cpp"
#include "mdiff_sort.cpp"
#include "mdiff_sort_omp.cpp"
#include "intro_sort.cpp"

typedef int key_type;

struct val_t
{
        key_type key;
        double d;
        val_t(const key_type& k = 0) : key(k), d(0) {}
        val_t(const val_t& v) : key(v.key), d(v.d) {}
        val_t& operator=(const val_t& v)
        {
                if (this != &v) new(this) val_t(v);
                return *this;
        }
        val_t& operator=(const key_type& k)
        {
                new(this) val_t(k);
                return *this;
        }

        bool operator==(const val_t& x) const { return key == x.key; }
};

struct cmp_t : public std::binary_function<val_t*,val_t*,bool>
{
        bool operator()(const val_t* x, const val_t* y) const
        {
                return x->key < y->key;
        }
};

struct sub_t : public std::binary_function<val_t*,val_t*,key_type>
{
        key_type operator()(const val_t* x, const val_t* y) const
        {
                return x->key - y->key;
        }
};

struct get_key_t : public std::unary_function<val_t*,key_type>
{
        key_type operator()(const val_t* x) const { return x->key; }
};

static void load(std::vector<val_t*>& a, std::vector<val_t>& b)
{
        typedef std::vector<val_t>::size_type size_type;
        const size_type N = b.size();
        a.resize(N);
        for (size_type i = 0; i < N; i++) a[i] = &b[i];
}

static void save(std::vector<val_t*>& a, std::vector<val_t>& b)
{
        typedef std::vector<val_t>::size_type size_type;
        const size_type N = a.size();
        std::vector<val_t> tmp(N);
        for(size_type i = 0; i < N; i++) tmp[i] = *a[i];
        b = tmp;
}

int main()
{
        using namespace std;
        static const int N = 1000000;
        static const int L = 30; 
        clock_t t1, t2, sum;
        vector<val_t> a(N), b, c;
        vector<val_t*> d;

        set_random(time(NULL));
        for(int i = 0; i < N; i++) a[i] = get_random();
        load(d, a);
        sort(d.begin(), d.end(), cmp_t());
        save(d, b);

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                comb_sort(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "comb_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                icomb_sort(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "icomb_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                quick_sort(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "quick_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                pquick_sort(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "pquick_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                pquick_sort2(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "pquick_sort2:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                merge_sort(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "merge_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                pmerge_sort(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "pmerge_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                pmerge_sort2(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "pmerge_sort2:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                mmerge_sort(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "mmerge_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                heap_sort1(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "heap_sort1:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                heap_sort2(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "heap_sort2:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                heap_sort3(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "heap_sort3:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                radix_sort<8,32>(d.begin(), d.end(), get_key_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "radix_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                msd_radix_sort<9,32>(d.begin(), d.end(), get_key_t(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "msd_radix_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                t1 = get_clock();
                load(d, a);
                diff_sort(d.begin(), d.end(), sub_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "diff_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                mdiff_sort(d.begin(), d.end(), sub_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "mdiff_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                mdiff_sort_omp(d.begin(), d.end(), sub_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "mdiff_sort_omp:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                intro_sort(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "intro_sort:\t" << setw(5) << millisecond(sum) / L;
        save(d, c);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                stable_sort(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "stable_sort:\t" << setw(5) << millisecond(sum) / L << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                load(d, a);
                t1 = get_clock();
                sort(d.begin(), d.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "sort:\t\t" << setw(5) << millisecond(sum) / L << endl;
}
