#define _CRT_RAND_S
#include <iostream>
#include <iomanip>
#include <functional>
#include <vector>
#include <list>
#include <cstdlib>
#include <ctime>

#include "benchmark.h"

#include "selection_sort.cpp"
#include "bubble_sort.cpp"
#include "insertion_sort.cpp"
#include "comb_sort.cpp"
#include "icomb_sort.cpp"
#include "quick_sort.cpp"
#include "pquick_sort.cpp"
#include "pquick_sort2.cpp"
#include "squick_sort.cpp"
#include "merge_sort.cpp"
#include "pmerge_sort.cpp"
#include "pmerge_sort2.cpp"
#include "mmerge_sort.cpp"
#include "oln_merge_sort.cpp"
#include "oln_merge_sort2.cpp"
#include "heap_sort1.cpp"
#include "heap_sort2.cpp"
#include "heap_sort3.cpp"
#include "radix_sort.cpp"
#include "msd_radix_sort.cpp"
#include "diff_sort.cpp"
#include "mdiff_sort.cpp"
#include "mdiff_sort_omp.cpp"
#include "intro_sort.cpp"

typedef int key_type;

struct val_t
{
        key_type key;
        int* p;
        val_t(const key_type& k = 0) : key(k), p(0) {}
        val_t(const val_t& v) : key(v.key), p(v.p) {}
        val_t& operator=(const val_t& v)
        {
                if (this != &v) new(this) val_t(v);
                return *this;
        }
        val_t& operator=(const key_type& k)
        {
                new(this) val_t(k);
                return *this;
        }

        bool operator==(const val_t& x) const { return key == x.key; }
};

struct cmp_t : public std::binary_function<val_t,val_t,bool>
{
        bool operator()(const val_t& x, const val_t& y) const
        {
                return x.key < y.key;
        }
};

struct sub_t : public std::binary_function<val_t,val_t,key_type>
{
        key_type operator()(const val_t& x, const val_t& y) const
        {
                return x.key - y.key;
        }
};

struct get_key_t : public std::unary_function<val_t,key_type>
{
        key_type operator()(const val_t& x) const { return x.key; }
};

int main()
{
        using namespace std;
        static const int N = 40;
        static const int L = 1000000; 
        clock_t t1, t2, sum;
        vector<val_t> a(N), b, c;

        set_random(time(NULL));
        for(int i = 0; i < N; i++) a[i] = get_random();
        b = a; sort(b.begin(), b.end(), cmp_t());

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                selection_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "selection_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                bubble_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "bubble_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                insertion_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "insertion_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                comb_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "comb_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                icomb_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "icomb_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                quick_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "quick_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                pquick_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "pquick_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                pquick_sort2(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "pquick_sort2:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                merge_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "merge_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                pmerge_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "pmerge_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                pmerge_sort2(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "pmerge_sort2:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                mmerge_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "mmerge_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                heap_sort1(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "heap_sort1:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                heap_sort2(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "heap_sort2:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                heap_sort3(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "heap_sort3:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                radix_sort<8,32>(c.begin(), c.end(), get_key_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "radix_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                msd_radix_sort<9,32>(c.begin(), c.end(), get_key_t(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "msd_radix_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                diff_sort(c.begin(), c.end(), sub_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "diff_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                mdiff_sort(c.begin(), c.end(), sub_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "mdiff_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                mdiff_sort_omp(c.begin(), c.end(), sub_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "mdiff_sort_omp:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                intro_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "intro_sort:\t" << setw(5) << millisecond(sum);
        if (b == c)
                cout << "\tOK" << endl;
        else
                cout << "\tNG" << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                stable_sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "stable_sort:\t" << setw(5) << millisecond(sum) << endl;

        sum = 0;
        for (int i = 0; i < L; i++) {
                c = a;
                t1 = get_clock();
                sort(c.begin(), c.end(), cmp_t());
                t2 = get_clock();
                sum += t2 - t1;
        }
        cout << "sort:\t\t" << setw(5) << millisecond(sum) << endl;
}
