#ifndef diff_sort_cpp
#define diff_sort_cpp

#include <iterator>
#include <functional>
#include <new>

template <class RAI, class S, class B>
static void diff_sort(int ld, RAI a0, RAI aN, S sub, B b);

template <class RAI, class S>
inline void diff_sort(RAI a0, RAI aN, S sub)
{
        typedef typename std::iterator_traits<RAI>::difference_type dif_t;
        typedef typename std::iterator_traits<RAI>::value_type val_t;
        const dif_t N = aN - a0;
        val_t* const b = new val_t[N];
        try { diff_sort(-1, a0, aN, sub, b); }
        catch (std::bad_alloc&) { delete[] b; throw; }
        delete[] b;
}

template <class RAI>
inline void diff_sort(RAI a0, RAI aN)
{
        typedef typename std::iterator_traits<RAI>::value_type val_t;
        diff_sort(a0, aN, std::minus<val_t>());
}

template <class RAI, class S, class B>
static void diff_sort(int ld, RAI a0, RAI aN, S sub, B b)
{
        typedef typename std::iterator_traits<RAI>::value_type val_t;
        typedef typename std::iterator_traits<RAI>::difference_type dif_t;
        typedef typename S::result_type key_type;
        static const int SIGS = 9;
        static const int KEYS = 1 << SIGS;
        const dif_t N = aN - a0;
        const RAI& a = a0;

        if (N < 40) {
                // Ĺ 40 꾮
                if (a0 == aN) return;
                RAI p = a0;
                for (++p; p != aN; ++p) {
                        const val_t tmp = *p;
                        RAI q, r; q = r = p;
                        if (sub(tmp, *a0) < 0)
                                while (a0 != q) { --r; *q = *r; q = r; }
                        else 
                                while (sub(tmp, *(--r)) < 0) { *q = *r; q = r; }
                        *q = tmp;
                }
                return;
        }
        val_t min_a, max_a;
        min_a = max_a = *a0;
        for (RAI ai = a0 + 1; ai < aN; ++ai) {
                const val_t& x = *ai;
                if (sub(x, min_a) < 0) min_a = x;
                else if (sub(max_a, x) < 0) max_a = x;
        }
        key_type d = sub(max_a, min_a);
        if (d == 0) return;
        if (ld < 0) for (ld = 0; d > 0; d >>= 1) ld++;
        ld = (ld > SIGS) ? ld - SIGS : 0;
        dif_t* const h = new dif_t[KEYS];
        for (int k = 0; k < KEYS; k++) h[k] = 0;
        B bi = b;
        for (RAI ai = a0; ai < aN; ++ai, ++bi) {
                const val_t& x = *ai;
                ++h[int(sub(x, min_a) >> ld)];
                *bi = x;
        }
        for (int k = 1; k < KEYS; k++) h[k] += h[k - 1];
        for (bi = b + N; bi > b;) {
                const val_t& x = *(--bi);
                const dif_t j = --h[int(sub(x, min_a) >> ld)];
                a[j] = x;
        }
        if (ld > 0) {
                for (int k = 0; k < KEYS - 1; k++)
                        diff_sort(ld, a0 + h[k], a0 + h[k + 1], sub, b);
                diff_sort(ld, a0 + h[KEYS - 1], aN, sub, b);
        }
        delete[] h;
}

#endif // diff_sort_cpp
