#ifndef merge_sort_cpp
#define merge_sort_cpp

#include <iterator>
#include <vector>
#include <functional>
#include <algorithm>

template <class FI, class C, class B, class D>
static void merge_sort(FI a0, FI aN, C cmp, B b0, D N);

template <class FI, class C>
inline void merge_sort(FI a0, FI aN, C cmp)
{
        typedef typename std::iterator_traits<FI>::value_type val_t;
        typedef typename std::iterator_traits<FI>::difference_type dif_t;
        const dif_t N = std::distance(a0, aN);
        const dif_t M = N >> 1;
        std::vector<val_t> b(M);
        merge_sort(a0, aN, cmp, b.begin(), N);
}

template <class FI>
inline void merge_sort(FI a0, FI aN)
{
        typedef typename std::iterator_traits<FI>::value_type val_t;
        merge_sort(a0, aN, std::less<val_t>());
}

template <class FI, class C, class B, class D>
static void merge_sort(FI a0, FI aN, C cmp, B b0, D N)
{
        if (N < 2) return;
        D M = N >> 1;
        FI aM = a0; std::advance(aM, M);
        merge_sort(a0, aM, cmp, b0, M);
        merge_sort(aM, aN, cmp, b0, N - M);
        B bi = b0, bM = std::copy(a0, aM, b0);
        FI aj = aM, ak = a0;
        while (bi != bM && aj != aN) {
                if (cmp(*aj, *bi))
                        *ak = *aj, ++ak, ++aj;
                else
                        *ak = *bi, ++ak, ++bi;
        }
        std::copy(bi, bM, ak);
}

#endif // merge_sort_cpp
