#ifndef my_thread_h
#define my_thread_h

#include <iostream>
#include <cstdlib>

#if defined(__WIN32__) || defined(_MSC_VER)

#include <windows.h>
#include <process.h>

typedef HANDLE thrd_handle_t;
typedef unsigned int thrd_return_t;
typedef thrd_return_t (WINAPI *thrd_func_t)(void *);

inline void start_thread(thrd_handle_t& th, thrd_func_t fp, void* arg)
{
        using namespace std;
        unsigned int thread_id;
        th = (thrd_handle_t) _beginthreadex(NULL, 0, fp, arg, 0, &thread_id);
        if (th == 0) { cerr << "error: _beginthreadex()" << endl; exit(1); }
}

inline void exit_thread() { _endthreadex(0); }

inline void wait_thread(thrd_handle_t& th)
{
        WaitForSingleObject(th, INFINITE);
        CloseHandle(th);
}

#else // PTHREAD

#include <pthread.h>

#define WINAPI
typedef pthread_t thrd_handle_t;
typedef void* thrd_return_t;
typedef thrd_return_t (*thrd_func_t)(void *);

inline void start_thread(thrd_handle_t& th, thrd_func_t fp, void* arg)
{
        using namespace std;
        int st = pthread_create(&th, NULL, fp, arg);
        if (st != 0) { cerr << "error: pthread_create()" << endl; exit(1); }
}

inline void exit_thread() { pthread_exit(0); }

inline void wait_thread(thrd_handle_t& th)
{
        using namespace std;
        int st = pthread_join(th, NULL);
        if (st != 0) { cerr << "error: pthread_join()" << endl; exit(1); }
}

#endif // __WIN32__

#endif // my_thread_h
