#include <cstdlib>
#include <cassert>
#include <search.h>
#include "sort_r.h"
#include "sort.hpp"
#include "bubble_sort.cpp"
#include <iostream>
#include <algorithm>
#include <functional>
#include "StopWatch.hpp"
using namespace std;

#define do_bench_sort(op) do {\
  StopWatch sw;\
\
  sw.start();\
  op;\
  sw.stop();\
  for (i=0; i<n-1; i++) if (x[i] < x[i+1]) break;\
  cout << left << setw(48) << #op << "\t" << ((i==n-1)?"":"!") << setw(10) << sw.get_ctime() << endl;\
  for (i=0; i<n; i++) x[i] = x0[i];\
} while (0)

#define do_bench_search(op,as) do {\
  StopWatch sw;\
\
  sw.start();\
  op;\
  sw.stop();\
  cout << left << setw(56) << #op << "\t" << ((as)?"":"!") << setw(10) << sw.get_ctime() << endl;\
} while (0)

#define T_RCMP(T) ((*(T *)a > *(T *)b) ? -1 : (*(T *)a < *(T *)b) ? 1 : 0)
int d_rcmp(const void *a, const void *b) { return T_RCMP(double); }
int d_rcmp_r(void *thunk, const void *a, const void *b) { return T_RCMP(double); }

void check_small()
{
  int n = 50000;
  double *x0 = new double[n], *x = new double[n];
  int i;

  for (i=0; i<n; i++)
    x[i] = x0[i] = (double)random()/((1<<31)-1);

  cout << "# number: " << n << endl;
  cout << left << setw(48) << "operation" << "\t" << setw(10) << "ctime(sec)" << endl;

  do_bench_sort(bubblesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(oddevensort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(insertsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(poddevensort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(bubble_sort(x, x+n, greater<double>()));
  do_bench_sort(foam_sort(x, x+n, greater<double>()));
  do_bench_sort(oddeven_sort(x, x+n, greater<double>()));

  do_bench_sort(parallel_oddeven_sort(x, x+n, greater<double>()));
}
void check()
{
  int n = 100000;
  double *x0 = new double[n], *x = new double[n];
  int i;
#if defined(__BLOCKS__)
  int (^d_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(double *)a > *(double *)b) ? -1 : (*(double *)a < *(double *)b) ? 1 : 0;
  };
#endif

  for (i=0; i<n; i++)
    x[i] = x0[i] = (double)random()/((1<<31)-1);

  cout << "# number: " << n << endl;
  cout << left << setw(48) << "operation" << "\t" << setw(10) << "ctime(sec)" << endl;

  do_bench_sort(combsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(shellsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(icombsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(pshellsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(shell_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_shell_sort(x, x+n, greater<double>()));

  do_bench_sort(heapsort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(heapsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(heapsort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(partialsort_r(x, n, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(partial_sort(x, x+n, x+n, greater<double>()));

  do_bench_sort(mergesort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(mergesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(mergesort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(timsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(pmergesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(ptimsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(stable_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_stable_sort(x, x+n, greater<double>()));

  do_bench_sort(qsort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(qsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(qsort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(introsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(psort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(pintrosort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_sort(x, x+n, greater<double>()));

  size_t m = n;
  double t, *f;
  bool b;

  t = x[random()%n];
  sort(x, x+n, greater<double>());

  cout << "# number: " << n << endl;
  cout << left << setw(56) << "operation" << "\t" << setw(10) << "ctime(sec)" << endl;

  do_bench_search(f = (double *)lfind(&t, x, &m, sizeof(x[0]), d_rcmp),			f && *f == t);
  do_bench_search(f = find(x, x+n, t),							f && *f == t);
  do_bench_search(f = (double *)bsearch(&t, x, n, sizeof(x[0]), d_rcmp),		f && *f == t);
  do_bench_search(f = (double *)lbsearch_r(&t, x, n, sizeof(x[0]), 0, d_rcmp_r),	f && *f == t);
  do_bench_search(f = (double *)ubsearch_r(&t, x, n, sizeof(x[0]), 0, d_rcmp_r),	f && *(f-1) == t);
  do_bench_search(f = (double *)bsearch_r(&t, x, n, sizeof(x[0]), 0, d_rcmp_r),		f && *f == t);
#if defined(__BLOCKS__)
  do_bench_search(f = (double *)bsearch_b(&t, x, n, sizeof(x[0]), d_rcmp_b),		f && *f == t);
#endif
  do_bench_search(f = lower_bound(x, x+n, t, greater<double>()),			f && *f == t);
  do_bench_search(f = upper_bound(x, x+n, t, greater<double>()),			f && *(f-1) == t);
  do_bench_search(b = binary_search(x, x+n, t, greater<double>()),			b);
}
int main()
{
  check_small();
  check();
  return 0;
}
