#include <cstdlib>
#include <cassert>
#include <search.h>
#include "sort_r.h"
#include "sort.hpp"
#include "bubble_sort.cpp"
#include <iostream>
#include <algorithm>
#include <functional>
#include "StopWatch.hpp"
using namespace std;

#define do_bench_sort(op) do {\
  StopWatch sw;\
\
  sw.start();\
  op;\
  sw.stop();\
  for (i=0; i<n-1; i++) if (x[ix[i]] < x[ix[i+1]]) break;\
  cout << left << setw(48) << #op << "\t" << ((i==n-1)?"":"!") << setw(10) << sw.get_ctime() << endl;\
  for (i=0; i<n; i++) ix[i] = i;\
} while (0)

#define do_bench_search(op,as) do {\
  StopWatch sw;\
\
  sw.start();\
  op;\
  sw.stop();\
  cout << left << setw(56) << #op << "\t" << ((as)?"":"!") << setw(10) << sw.get_ctime() << endl;\
} while (0)

#define ARI_T_RCMP_R(T) ((((T *)thunk)[*(int *)a] > ((T *)thunk)[*(int *)b]) ? -1 : (((T *)thunk)[*(int *)a] < ((T *)thunk)[*(int *)b]) ? 1 : 0)
int ari_d_rcmp_r(void *thunk, const void *a, const void *b)	{ return ARI_T_RCMP_R(double); }

template <typename T>
struct ari_rcomp { ari_rcomp(T *p):v(p){} bool operator()(const int &a, const int &b) { return v[a] > v[b]; } private: T *v; };

#define ARI_T_RE_R(T) ((*(T *)k > ((T *)thunk)[*(int *)i]) ? -1 : (*(T *)k < ((T *)thunk)[*(int *)i]) ? 1 : 0)
int ari_d_re_r(void *thunk, const void *k, const void *i) { return ARI_T_RE_R(double); }

template <typename T>
struct ari_rpred { ari_rpred(T *p, T *q):v(p),t(q){} bool operator()(const int &a) { return v[a] == *t; } private: T *v, *t; };

template <typename T>
struct ari_rkomp { ari_rkomp(T *p):v(p){} bool operator()(const int &i, const T &k) { return v[i] > k; } private: T *v; };

template <typename T>
struct ari_rkowp { ari_rkowp(T *p):v(p){} bool operator()(const T &k, const int &i) { return k > v[i]; } private: T *v; };

void check_small()
{
  int n = 50000;
  double *x0 = new double[n], *x = new double[n];
  int i, *ix = new int[n];
  typedef ari_rcomp<double> ari_d_rcomp;

  for (i=0; i<n; i++)
    x[i] = x0[ix[i] = i] = (double)random()/((1<<31)-1);

  cout << "# number: " << n << endl;
  cout << left << setw(48) << "operation" << "\t" << setw(10) << "ctime(sec)" << endl;

  do_bench_sort(bubblesort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
  do_bench_sort(oddevensort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
  do_bench_sort(insertsort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));

  do_bench_sort(poddevensort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));

  do_bench_sort(bubble_sort(ix, ix+n, ari_d_rcomp(x)));
  do_bench_sort(foam_sort(ix, ix+n, ari_d_rcomp(x)));
  do_bench_sort(oddeven_sort(ix, ix+n, ari_d_rcomp(x)));

  do_bench_sort(parallel_oddeven_sort(ix, ix+n, ari_d_rcomp(x)));
}
void check()
{
  int n = 100000;
  double *x0 = new double[n], *x = new double[n];
  int i, *ix = new int[n];
  typedef ari_rcomp<double> ari_d_rcomp;
  typedef ari_rpred<double> ari_d_rpred;
  typedef ari_rkomp<double> ari_d_rkomp;
  typedef ari_rkowp<double> ari_d_rkowp;
#if defined(__BLOCKS__)
  int (^ari_d_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] > x[*(int *)b]) ? -1 : (x[*(int *)a] < x[*(int *)b]) ? 1 : 0;
  };
  int (^ari_d_re_b)(const void *k, const void *i) = ^(const void *k, const void *i) {
    return (*(double *)k > x0[*(int *)i]) ? -1 : (*(double *)k < x0[*(int *)i]) ? 1 : 0;
  };
#endif

  for (i=0; i<n; i++)
    x[i] = x0[ix[i] = i] = (double)random()/((1<<31)-1);

  cout << "# number: " << n << endl;
  cout << left << setw(48) << "operation" << "\t" << setw(10) << "ctime(sec)" << endl;

  do_bench_sort(combsort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
  do_bench_sort(shellsort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
  do_bench_sort(icombsort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));

  do_bench_sort(pshellsort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));

  do_bench_sort(shell_sort(ix, ix+n, ari_d_rcomp(x)));
  do_bench_sort(parallel_shell_sort(ix, ix+n, ari_d_rcomp(x)));

  do_bench_sort(heapsort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(heapsort_b(ix, n, sizeof(ix[0]), ari_d_rcmp_b));
#endif
  do_bench_sort(partialsort_r(ix, n, n, sizeof(ix[0]), x, ari_d_rcmp_r));
  do_bench_sort(partial_sort(ix, ix+n, ix+n, ari_d_rcomp(x)));

  do_bench_sort(mergesort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(mergesort_b(ix, n, sizeof(ix[0]), ari_d_rcmp_b));
#endif
  do_bench_sort(timsort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
  do_bench_sort(pmergesort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
  do_bench_sort(ptimsort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
  do_bench_sort(stable_sort(ix, ix+n, ari_d_rcomp(x)));
  do_bench_sort(parallel_stable_sort(ix, ix+n, ari_d_rcomp(x)));

  do_bench_sort(qsort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(qsort_b(ix, n, sizeof(ix[0]), ari_d_rcmp_b));
#endif
  do_bench_sort(introsort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(psort_b(ix, n, sizeof(ix[0]), ari_d_rcmp_b));
#endif
  do_bench_sort(pintrosort_r(ix, n, sizeof(ix[0]), x, ari_d_rcmp_r));
  do_bench_sort(sort(ix, ix+n, ari_d_rcomp(x)));
  do_bench_sort(parallel_sort(ix, ix+n, ari_d_rcomp(x)));

  size_t m = n;
  double t;
  int *g, I;

  t = x0[I=random()%n];
  sort(ix, ix+n, ari_d_rcomp(x0));

  cout << "# number: " << n << endl;
  cout << left << setw(56) << "operation" << "\t" << setw(10) << "ctime(sec)" << endl;

  do_bench_search(g = find_if(ix, ix+n, ari_d_rpred(x0,&t)),				g && x0[*g] == t);
  do_bench_search(g = (int *)lfind_r(&t, ix, &m, sizeof(ix[0]), x0, ari_d_re_r),	g && x0[*g] == t);
  do_bench_search(g = (int *)lbsearch_r(&t, ix, n, sizeof(ix[0]), x0, ari_d_re_r),	g && x0[*g] == t);
  do_bench_search(g = (int *)ubsearch_r(&t, ix, n, sizeof(ix[0]), x0, ari_d_re_r),	g && x0[*(g-1)] == t);
  do_bench_search(g = (int *)bsearch_r(&t, ix, n, sizeof(ix[0]), x0, ari_d_re_r),	g && x0[*g] == t);
#if defined(__BLOCKS__)
  do_bench_search(g = (int *)bsearch_b(&t, ix, n, sizeof(ix[0]), ari_d_re_b),		g && x0[*g] == t);
#endif
  do_bench_search(g = lower_bound(ix, ix+n, t, ari_d_rkomp(x0)),			g && x0[*g] == t);
  do_bench_search(g = upper_bound(ix, ix+n, t, ari_d_rkowp(x0)),			g && x0[*(g-1)] == t);
}
int main()
{
  check_small();
  check();
  return 0;
}
