#include <cstdlib>
#include <cassert>
#include <search.h>
#include "sort_r.h"
#include "sort.hpp"
#include "bubble_sort.cpp"
#include "quick_sort.cpp"
#include "quick_sort2.cpp"
#include "pquick_sort.cpp"
#include "pquick_sort2.cpp"
#include "intro_sort.cpp"
#include <iostream>
#include <algorithm>
#include <functional>
#include "StopWatch.hpp"
#define QSORT_ONLY
using namespace std;

#define do_bench_sort(op) do {\
  StopWatch sw;\
\
  sw.start();\
  op;\
  sw.stop();\
  for (i=0; i<n-1; i++) if (x[i] < x[i+1]) break;\
  cout << left << setw(48) << #op << "\t" << ((i==n-1)?"":"!") << setw(10) << sw.get_ctime() << endl;\
  for (i=0; i<n; i++) x[i] = x0[i];\
} while (0)

#define T_RCMP(T) ((*(T *)a > *(T *)b) ? -1 : (*(T *)a < *(T *)b) ? 1 : 0)
int d_rcmp(const void *a, const void *b) { return T_RCMP(double); }
int d_rcmp_r(void *thunk, const void *a, const void *b) { return T_RCMP(double); }

void check_random()
{
  int n = 1000000;
  double *x0 = new double[n], *x = new double[n];
  int i;
#if defined(__BLOCKS__)
  int (^d_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(double *)a > *(double *)b) ? -1 : (*(double *)a < *(double *)b) ? 1 : 0;
  };
#endif

  for (i=0; i<n; i++)
    x[i] = x0[i] = (double)random()/((1<<31)-1);

  cout << "# number: " << n << ",\t" << "type of sequence: random" << endl;
  cout << left << setw(48) << "operation" << "\t" << setw(10) << "ctime(sec)" << endl;

#if !defined(QSORT_ONLY)
  do_bench_sort(combsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(shellsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(icombsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(pshellsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(shell_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_shell_sort(x, x+n, greater<double>()));

  do_bench_sort(heapsort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(heapsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(heapsort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(partialsort_r(x, n, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(partial_sort(x, x+n, x+n, greater<double>()));

  do_bench_sort(mergesort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(mergesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(mergesort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(timsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(pmergesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(ptimsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(stable_sort(x, x+n, greater<double>()));
  do_bench_sort(sequential_stable_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_stable_sort(x, x+n, greater<double>()));
#endif

  do_bench_sort(qsort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(qsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(qsort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(introsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(psort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(pintrosort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(quick_sort(x, x+n, greater<double>()));
  //do_bench_sort(quick_sort2(x, x+n, greater<double>()));
  do_bench_sort(pquick_sort(x, x+n, greater<double>()));
  do_bench_sort(pquick_sort2(x, x+n, greater<double>()));
  do_bench_sort(intro_sort(x, x+n, greater<double>()));
  do_bench_sort(sort(x, x+n, greater<double>()));
  do_bench_sort(sequential_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_sort(x, x+n, greater<double>()));
}
void check_unrandom()
{
  int n = 1000000;
  double *x0 = new double[n], *x = new double[n];
  int i, j;
#if defined(__BLOCKS__)
  int (^d_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(double *)a > *(double *)b) ? -1 : (*(double *)a < *(double *)b) ? 1 : 0;
  };
#endif

  for (i=0; i<n; i+=1000) {
    x[i] = x0[i] = (double)random()/((1<<31)-1);
    for (j=1; j<1000 && i+j<n; j++)
      x[i+j] = x0[i+j] = x0[i];
  }

  cout << "# number: " << n << ",\t" << "type of sequence: unrandom" << endl;
  cout << left << setw(48) << "operation" << "\t" << setw(10) << "ctime(sec)" << endl;

#if !defined(QSORT_ONLY)
  do_bench_sort(combsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(shellsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(icombsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(pshellsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(shell_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_shell_sort(x, x+n, greater<double>()));

  do_bench_sort(heapsort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(heapsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(heapsort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(partialsort_r(x, n, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(partial_sort(x, x+n, x+n, greater<double>()));

  do_bench_sort(mergesort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(mergesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(mergesort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(timsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(pmergesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(ptimsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(stable_sort(x, x+n, greater<double>()));
  do_bench_sort(sequential_stable_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_stable_sort(x, x+n, greater<double>()));
#endif

  do_bench_sort(qsort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(qsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(qsort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(introsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(psort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(pintrosort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(quick_sort(x, x+n, greater<double>()));
  //do_bench_sort(quick_sort2(x, x+n, greater<double>()));
  do_bench_sort(pquick_sort(x, x+n, greater<double>()));
  do_bench_sort(pquick_sort2(x, x+n, greater<double>()));
  do_bench_sort(intro_sort(x, x+n, greater<double>()));
  do_bench_sort(sort(x, x+n, greater<double>()));
  do_bench_sort(sequential_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_sort(x, x+n, greater<double>()));
}
void check_sorted()
{
  int n = 1000000;
  double *x0 = new double[n], *x = new double[n];
  int i;
#if defined(__BLOCKS__)
  int (^d_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(double *)a > *(double *)b) ? -1 : (*(double *)a < *(double *)b) ? 1 : 0;
  };
#endif

  for (i=0; i<n; i++)
    x0[i] = (double)random()/((1<<31)-1);
  sort(x0, x0+n, greater<double>());
  for (i=0; i<n; i++)
    x[i] = x0[i];

  cout << "# number: " << n << ",\t" << "type of sequence: sorted" << endl;
  cout << left << setw(48) << "operation" << "\t" << setw(10) << "ctime(sec)" << endl;

#if !defined(QSORT_ONLY)
  do_bench_sort(combsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(shellsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(icombsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(pshellsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(shell_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_shell_sort(x, x+n, greater<double>()));

  do_bench_sort(heapsort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(heapsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(heapsort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(partialsort_r(x, n, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(partial_sort(x, x+n, x+n, greater<double>()));

  do_bench_sort(mergesort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(mergesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(mergesort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(timsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(pmergesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(ptimsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(stable_sort(x, x+n, greater<double>()));
  do_bench_sort(sequential_stable_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_stable_sort(x, x+n, greater<double>()));
#endif

  do_bench_sort(qsort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(qsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(qsort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(introsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(psort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(pintrosort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  //  do_bench_sort(quick_sort(x, x+n, greater<double>()));
  //do_bench_sort(quick_sort2(x, x+n, greater<double>()));
  do_bench_sort(pquick_sort(x, x+n, greater<double>()));
  do_bench_sort(pquick_sort2(x, x+n, greater<double>()));
  do_bench_sort(intro_sort(x, x+n, greater<double>()));
  do_bench_sort(sort(x, x+n, greater<double>()));
  do_bench_sort(sequential_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_sort(x, x+n, greater<double>()));
}
void check_med3killer()
{
  int n = 1000000;
  double *x0 = new double[n], *x = new double[n];
  int i;
#if defined(__BLOCKS__)
  int (^d_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(double *)a > *(double *)b) ? -1 : (*(double *)a < *(double *)b) ? 1 : 0;
  };
#endif

  /*
    Median-of-3 killer sequences:

    n=8:  1 5 3 7|2 4 6 8
    n=12: 1 7 3 9 5 11|2 4 6 8 10 12

    n=2k: 1 k+1 3 k+3 5 ... 2k-3 k-1 2k-1|2 4 6 ... 2k-2 2k
  */
  for (i=1; i<=n; i++)
    x0[i-1] = (i<=n/2) ? ((i%2)?i:(n/2+i-1)) : ((i-n/2)*2);
  for (i=0; i<n; i++)
    x[i] = x0[i];

  cout << "# number: " << n << ",\t" << "type of sequence: med3killer" << endl;
  cout << left << setw(48) << "operation" << "\t" << setw(10) << "ctime(sec)" << endl;

#if !defined(QSORT_ONLY)
  do_bench_sort(combsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(shellsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(icombsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(pshellsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));

  do_bench_sort(shell_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_shell_sort(x, x+n, greater<double>()));

  do_bench_sort(heapsort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(heapsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(heapsort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(partialsort_r(x, n, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(partial_sort(x, x+n, x+n, greater<double>()));

  do_bench_sort(mergesort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(mergesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(mergesort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(timsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(pmergesort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(ptimsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(stable_sort(x, x+n, greater<double>()));
  do_bench_sort(sequential_stable_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_stable_sort(x, x+n, greater<double>()));
#endif

  do_bench_sort(qsort(x, n, sizeof(x[0]), d_rcmp));
  do_bench_sort(qsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(qsort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(introsort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
#if defined(__BLOCKS__)
  do_bench_sort(psort_b(x, n, sizeof(x[0]), d_rcmp_b));
#endif
  do_bench_sort(pintrosort_r(x, n, sizeof(x[0]), NULL, d_rcmp_r));
  do_bench_sort(quick_sort(x, x+n, greater<double>()));
  //do_bench_sort(quick_sort2(x, x+n, greater<double>()));
  do_bench_sort(pquick_sort(x, x+n, greater<double>()));
  do_bench_sort(pquick_sort2(x, x+n, greater<double>()));
  do_bench_sort(intro_sort(x, x+n, greater<double>()));
  do_bench_sort(sort(x, x+n, greater<double>()));
  do_bench_sort(sequential_sort(x, x+n, greater<double>()));
  do_bench_sort(parallel_sort(x, x+n, greater<double>()));
}
int main()
{
  check_random();
  check_unrandom();
  check_sorted();
  check_med3killer();
  return 0;
}
