#ifndef benchmark_h
#define benchmark_h

#define _CRT_RAND_S
#include <cstdlib>
#include <cmath>

#if defined(__WIN32__) || defined(_MSC_VER)
#include <ctime>
#else
#include <sys/times.h>
#include <unistd.h>
#endif

inline void set_random(unsigned int s)
{
#ifdef USE_RANDOM
        srandom(s);
#else
        srand(s);
#endif // USE_RADDOM
}

inline long get_random()
{
#ifdef _MSC_VER
        unsigned int n;
        rand_s(&n);
        return (n >> 1) & 0x7fffffff;
#else
#ifdef USE_RANDOM
        return random() & 0x7fffffff;
#else
#ifdef __WIN32__
        return ((rand() << 16) + rand()) & 0x7fffffff;
#else
        return rand() & 0x7fffffff;
#endif // __WIN32__
#endif // USE_RADDOM
#endif // _MSC_VER
}


#if defined(__WIN32__) || defined(_MSC_VER)

inline clock_t get_clock() { return clock(); }

inline int millisecond(const clock_t& t)
{
        const double ms = 1000 * double(t) / CLOCKS_PER_SEC;
        return static_cast<int>(floor(ms + 0.5));
}

#else

inline clock_t get_clock() { struct tms t; return times(&t); }

inline int millisecond(const clock_t& t)
{
        static const clock_t CPS = sysconf(_SC_CLK_TCK);
        const double ms = 1000 * double(t) / CPS;
        return static_cast<int>(floor(ms + 0.5));
}

#endif

#endif // benchmark_h
