#ifndef pmerge_sort_cpp
#define pmerge_sort_cpp

#include <iterator>
#include <vector>
#include <functional>
#include <algorithm>
#include "insertion_sort.cpp"

template <class BI, class C, class B, class D>
static void pmerge_sort(BI a0, BI aN, C cmp, B b0, D N);

template <class BI, class C>
inline void pmerge_sort(BI a0, BI aN, C cmp)
{
        typedef typename std::iterator_traits<BI>::value_type val_t;
        typedef typename std::iterator_traits<BI>::difference_type dif_t;
        const dif_t N = std::distance(a0, aN);
        const dif_t M = N >> 1;
        std::vector<val_t> b(M);
        pmerge_sort(a0, aN, cmp, b.begin(), N);
}

template <class BI>
inline void pmerge_sort(BI a0, BI aN)
{
        typedef typename std::iterator_traits<BI>::value_type val_t;
        pmerge_sort(a0, aN, std::less<val_t>());
}

template <class BI, class C, class B, class D>
static void pmerge_sort(BI a0, BI aN, C cmp, B b0, D N)
{
        if (N < 34) { insertion_sort(a0, aN, cmp); return; }
        D M = N >> 1;
        BI aM = a0; std::advance(aM, M);
        pmerge_sort(a0, aM, cmp, b0, M);
        pmerge_sort(aM, aN, cmp, b0, N - M);
        B bi = b0, bM = std::copy(a0, aM, b0);
        BI aj = aM, ak = a0;
        while (bi != bM && aj != aN) {
                if (cmp(*aj, *bi))
                        *ak = *aj, ++ak, ++aj;
                else
                        *ak = *bi, ++ak, ++bi;
        }
        std::copy(bi, bM, ak);
}

#endif // pmerge_sort_cpp
